import networkx as nx
import functions as fx

### maximum_game_lenght is a variable that we use to track how many
# turns are played at most until FP wins.
maximum_game_length = []

### To improve computation time, we create a list of all non-isomorphic 
# board states at the end of stage 2 together with the board states
# at the beginning of the special stages 5 and 6.

# In order to do this we introduce unique_boards which saves all possible
# board states up to stage 2, 5 or 6.
# This takes at most 12 moves, hence the key values of unique_boards.
unique_boards = {2: [], 4: [], 6: [], 8: [], 10: [], 12: []}

### The algorithm that produces the board states up to stages 2, 5 and 6
# is essentially the operation of fx.move(), except that the
# stages are saved to unique_boards.
# We cannot use fx.move() directly since it is a recursive function
# and we only want to go at most 6 turns deep, not until the end of the game.
for turn in [0,2,4,6,8,10]:
    if turn == 0:
        # In the first turn we call FP_edge() and add that edge to the board.
        board = nx.Graph()
        FP_info = fx.FP_edge(board)
        board.add_edges_from([FP_info['move']], colour = 'FP', added_in_step = 0 )
        # We add two nodes with degree 0.
        board.add_nodes_from([100,101]) 
        # Afterwards we add all possible edges for SP s.t. the resulting board 
        # is non-isomorphic to the ones already seen.
        for edge in list(nx.non_edges(board)):
            board.add_edges_from([edge], colour = 'SP', added_in_step = 1)
            for vertex in list(nx.isolates(board)):
                board.remove_node(vertex)
            not_in_list = True
            for new_pos in list(unique_boards[turn+2]):
                saved_board = nx.Graph()
                saved_board.add_edges_from(new_pos['board'])
                if nx.is_isomorphic(saved_board,board,edge_match = fx.em):
                    not_in_list = False
                    break
            if not_in_list:
                unique_boards[turn+2].append({'board': list(board.edges(data=True)),'stage': 1})
            board.remove_edge(*edge)
    else:
        # In a later turn, the board states of the previous turn are considered.
        for old_board_state in unique_boards[turn]:
            # The board is created with the previous state.
            board = nx.Graph()
            board.add_edges_from(old_board_state['board'])
            # If the stage of the previous state was already larger than 3,
            # then we just add it to the current state.
            if old_board_state['stage'] in [3,4,5,6]:
                not_in_list = True
                for new_pos in unique_boards[turn+2]:
                    saved_board = nx.Graph()
                    saved_board.add_edges_from(new_pos['board'])
                    if nx.is_isomorphic(saved_board,board,edge_match = fx.em):
                        not_in_list = False
                        break
                if not_in_list:
                    unique_boards[turn+2].append({'board': list(board.edges(data=True)), 'stage': int(old_board_state['stage'])})
                continue
            # Otherwise we are in stage 1 or 2 and check which edge FP wants to add.
            FP_info = fx.FP_edge(board, old_board_state['stage'])
            stage = FP_info['alg_stage']
            # If FP_edge() says that FP wants to switch to stage 5 or 6,
            # we do not add the edge and instead add the board to the list.
            if old_board_state['stage'] not in [5,6] and FP_info['alg_stage'] in [5,6]:
                unique_boards[turn+2].append({'board': list(board.edges(data=True)),'stage': int(stage)})
                continue
            # Otherwise we really are in stage 1 or 2 and add FPs edge to the board.
            board.add_edges_from([FP_info['move']], colour = 'FP', added_in_step = turn )
            # First we make sure that there are exactly 2 vertices with no incident edges.
            # This ensures that, even though there are infinitely many vertices on the board,
            # we check all possible moves for SP up to isomorphism.
            for i in range(2-nx.number_of_isolates(board)):
                board.add_node(max(board.nodes()) + 1)
            # We create a list of all possible moves for SP, 
            # which is any unclaimed edge on the current board.
            for edge in list(nx.non_edges(board)):
                # We add SPs edge to the board.
                board.add_edges_from([edge], colour = 'SP', added_in_step = turn + 1)
                # If there is a threat by FP, then we simply delete SPs edge and
                # continue with the next possible edge for SP.
                if fx.threat_checker(board,'FP'):
                    board.remove_edge(*edge)
                    continue
                # We delete the vertices of degree 0 from the board before continuing to the next move.
                for vertex in list(nx.isolates(board)):
                    board.remove_node(vertex)
                # Next, we check whether the graph is already in the list for this turn.
                not_in_list = True
                for new_pos in unique_boards[turn+2]:
                    saved_board = nx.Graph()
                    saved_board.add_edges_from(new_pos['board'])
                    if nx.is_isomorphic(saved_board,board,edge_match = fx.em):
                        not_in_list = False
                        break
                # If it is not in the list, we add it to the list.
                if not_in_list:
                    unique_boards[turn+2].append({'board': list(board.edges(data=True)),'stage': int(stage)})
                # Afterwards, we need to delete SPs edge from the board in this case,
                # so we can add the next one during the loop.
                board.remove_edge(*edge)



# With this setup we want to check every possible course of the game, 
# starting at each of the unique position we found in the last step.
unique_graphs = unique_boards[12]
for unique_state in unique_graphs:
    # For each of the non-isomorphic board states after stage 2,
    # we create the board and pass the board edges, the nodes and
    # the stage to the move() function that recursively checks all possible
    # courses of the game that may result from this position.
    stage = unique_state['stage']
    edgeset = unique_state['board']
    aux_board = nx.Graph()
    aux_board.add_edges_from(edgeset)
    fx.move(list(edgeset), list(aux_board.nodes(data=True)), int(stage), )




# Lastly, we check how long the maximum game length was
print('There are at most {} moves made until the first player wins.'.format(fx.maximum_game_length_number_of_moves))
print('That means that the first player makes {:.0f} moves until she wins.'.format((fx.maximum_game_length_number_of_moves+1)/2))
# print('A board where the most moves were made is {}'.format(fx.maximum_game_length_board.edges(data=True)))

# and which size a board must have at least s.t. FP can execute her strategy
print('FP needs a board size of at least {} vertices to execute her strategy.'.format(fx.number_of_required_vertices))